/** Mode: -C++- 
 *
 * binpacker 1.0.2
 * 
 * (Time-stamp: <2004-10-18 15:36:37 boris@folgmann.de>)
 *
 * (C) 2004 by Folgmann IT-Consulting
 *
 */

// INCLUDES
#include <iostream>
#include <string>
#include <set>
#include <algorithm>
#include <sys/stat.h>

using namespace std;

class File
{
public:
  string m_name;
  off_t  m_size;

  File(string name, off_t size) { m_name = name; m_size = size; };
  bool operator<(const File &rhs) const
     { 
	if (m_size == rhs.m_size)
  	  return m_name > rhs.m_name;
	else
	  return m_size < rhs.m_size;
     };
  bool operator>(const File &rhs) const
     { 
	if (m_size == rhs.m_size)
  	  return m_name < rhs.m_name;
	else
	  return m_size > rhs.m_size;
     };
  void print(ostream &out) const { out << m_name << ": " << m_size << endl; };
  void print(void) const { cout << m_name << ": " << m_size << endl; };
};

// Set of Files, sorted in descending size order
// Files of the same size are sorted alphabetically
// Use a set<> instead of a multiset<> if multiple entries
// with same filename and size should be ignored
typedef multiset<File,greater<File> > FileSet;

// GLOBALS
FileSet Files;
off_t BinSize;

// FUNCTIONS
void PrintUsage()
{
  cerr << "Usage: binpacker binsize [files]" << endl;
  cerr << "Examples: binpacker 10000 a.png b.png" << endl;
  cerr << "          binpacker 52k *" << endl;
  cerr << "          binpacker 640m *.mp3" << endl;
  cerr << "          binpacker 4g *.mpg" << endl;
  exit(-1);
}

void PrintFiles()
{
  FileSet::iterator pos;
  for (pos = Files.begin(); pos != Files.end(); pos++)
    {
      pos->print(cout);
    }
}

void PrintFiles2()
{
  //  for_each(Files.begin(), Files.end(), print);
}

void CalcStatistics()
{
  FileSet::iterator pos;
  off_t total_sum = 0;

  for (pos = Files.begin(); pos != Files.end(); pos++)
    {
      total_sum += pos->m_size;
    }

  cout << endl 
       << "Bin size:\t"  << BinSize                               << endl;
  cout << "Total Sum:\t" << total_sum                             << endl; 
  cout << "Max bins:\t"  << 1 + ((total_sum-1) / (BinSize/2 + 1)) << endl; 
  cout << "Min bins:\t"  << 1 + ((total_sum-1) / BinSize)         << endl; 
}

void BinPacking()
{
  FileSet::iterator pos = Files.begin();

  for (int bucket = 1; !Files.empty(); bucket++)
    {
      cout << endl << "Bin #" << bucket << endl;
      off_t sum = 0;

      for (pos = Files.lower_bound(File("", BinSize - sum));
	   pos != Files.end();
	   pos = Files.lower_bound(File("", BinSize - sum)))
	{ // pos points to biggest file which still fits in the bucket
	  sum += pos->m_size;
	  pos->print(cout);
	  Files.erase(pos);
	}

      cout << "Uses " << sum << " of " << BinSize
	   << " bytes = " << (sum*100)/BinSize << "%" << endl;
    }
}


void GetFileSizes(char **argv)
{
  struct stat buf;
  char* filename;
  bool first_large = true;

  while ((filename = *(++argv)))
    {
      if (0 == stat(filename, &buf))
	{
	  if (S_ISDIR(buf.st_mode))
	    {
	      cerr << "Ignored directory " << filename << endl;
	      continue;
	    }
	  if (buf.st_size <= BinSize)
	    Files.insert(File(filename, buf.st_size)); 
	  else
	    {
	      if (first_large)
		{
		  first_large = false;
		  cout << "Ignored files which do not fit in bin:" << endl;
		}
	      cout << filename << ": " << buf.st_size << endl;
	    }
	}
      else
	{
	  cerr << "Could not stat file " << filename << endl;
	  exit(-1);
	}
    }
}

void GetBinSize(string binSize)
{
  char modifier = toupper(binSize[binSize.length()-1]);
  off_t mult;

  switch (modifier)
    {
      case 'K': mult = 1024;                break;
      case 'M': mult = 1024*1024;           break;
      case 'G': mult = 1024*1024*1024;      break;
	//case 'T': mult = 1024*1024*1024*1024; break; //overflow??
      default:  mult = 1;                   break; // no modifier 
    }

  BinSize = strtoll(binSize.c_str(), NULL, 10) * mult;
  if (BinSize == 0) // to prevent division by zero later
    PrintUsage();
}

int main(int argc, char **argv)
{
  if (argc < 2)
    PrintUsage();

  GetBinSize(*++argv);
  GetFileSizes(argv);
  //PrintFiles();
  CalcStatistics();
  BinPacking();
}
